"""
Options and enums for ShrinkToObjectCenters module
"""

# Note: ShrinkToObjectCenters is a simple morphological operation module that doesn't require
# complex enums. It has no settings beyond basic "Object Processing".
